import hashlib
from os.path import isfile

from rdflib import Graph
from rdflib.term import URIRef, Literal, BNode, Variable


hashlib_methods = {
    "md5": hashlib.md5,
    "sha1": hashlib.sha1,
    "sha224": hashlib.sha224,
    "sha256": hashlib.sha256,
    "sha384": hashlib.sha384,
    "sha512": hashlib.sha512,
    "sha3_224": hashlib.sha3_224,
    "sha3_256": hashlib.sha3_256,
    "sha3_384": hashlib.sha3_384,
    "sha3_512": hashlib.sha3_512,
    # "shake_128": hashlib.shake_128,
    # "shake_256": hashlib.shake_256,
    "blake2b": hashlib.blake2b,
    "blake2s": hashlib.blake2s,
}


def hash_string(s: str, method="sha256", hashlib_options={}) -> str:
    """Hash a Python string with a given

    Args:
        s (str): _description_
        method (str, optional): _description_. Defaults to "sha256".

    Raises:
        ValueError: _description_

    Returns:
        str: Hexadecimal string representation of hash.
    """
    if method in hashlib_methods:
        return hashlib_methods[method](s.encode("utf-8"), **hashlib_options).hexdigest()
    else:
        raise ValueError(f"Invalid hashing method: {method}")


def convert_data_to_graph(data, format: str = None) -> Graph:
    """Convert data provided to an rdflib.Graph.

    Args:
        data (_type_): Data representing RDF triples.
        format (str, optional): Format of data. Defaults to None.

    Returns:
        Graph: rdflib.Graph.
    """
    type_data = type(data)

    if type_data == Graph:
        return data

    elif type_data == str:
        # If is a file, parse file and always let rdflib determine format.
        if isfile(data):
            return Graph().parse(data)
        # If not a file, assume it's a string of RDF triples.
        else:
            return Graph().parse(data=data, format=format)

    elif type_data == list:
        graph = Graph()
        # Concatenate graphs together for each item in list.
        for item in data:
            graph = graph + convert_data_to_graph(item, format)
        return graph


def rdf_term_to_id(term: URIRef or Literal or BNode or Variable) -> str:
    """Convert rdflib.term to a resolvable identifier. This is useful for hashing.

    Args:
        term (URIRef or Literal or BNode or Variable): rdflib.term.

    Returns:
        str: String identifier for rdflib.term.
    """
    type_t = type(term)

    if type_t == URIRef:
        return term.n3()
    elif type_t == Literal:
        return (
            # If is an 'xsd:string', manually form id.
            f'"{term.value}"^^<http://www.w3.org/2001/XMLSchema#string>'
            if term.datatype == None
            # Else use '.n3()' to form id.
            else term.n3()
        )
    else:
        raise ValueError(
            "RDF Term type cannot be converted to a resolvable identifier: "
            + str(type_t)
        )



from os import getcwd, path, listdir
from pathlib import Path

from rdflib import Graph
import pytest

from package.hash import rdfhash
from package.logger import logger

hash_methods = ["sha256"]

script_path = Path(__file__).parent.absolute()
examples_path = path.join(script_path, "..", "examples")


def test__hash_examples(force_write=False):
    """Test hashing files in './examples'.

    Args:
        force_write (bool, optional): If True, forces writing hash
            result to file './examples/hashed'. Defaults to True.
    """
    # Get list of all items (files+directories) in './examples'.
    test_files = set(Path(v).name for v in listdir(examples_path))

    results = {"not_found": [], "successful": [], "failed": [], "wrote": []}

    # Test each file in './examples'.
    # -------------------------------

    # Iterate over each file in './examples'.
    for file in test_files:
        file_path = path.join(getcwd(), "examples", file)

        # If is not a file, skip.
        if not path.isfile(file_path):
            continue

        # Iterate over hash methods defined above.
        for method in hash_methods:
            hash_file_path = path.join(
                getcwd(),
                "examples",
                "hashed",
                f"{path.splitext(file)[0]}__{method}.ttl",
            )

            # Generate hash of blank nodes in example file.
            graph = rdfhash(file_path, method=method)

            graph_actual = (
                None
                if not path.isfile(hash_file_path)
                else Graph().parse(hash_file_path)
            )

            # If hash file does not exist, continue.
            if graph_actual == None:
                logger.warning(f"Cannot find hash file at path: {hash_file_path}")
                results["not_found"].append(hash_file_path)

            # Check to see that both graphs are the exact same.
            elif graph.isomorphic(graph_actual):
                logger.info(
                    "Successfully verified hash against file: "
                    f"'{file}' <-> '{hash_file_path}' ({method})"
                )
                results["successful"].append(file_path)

            # If the hash is not correct, append to 'failed' and continue.
            else:
                logger.error(
                    "Mismatch between calculated hash and file: "
                    f"'{file}' -> '{hash_file_path}' ({method}) "
                )
                results["failed"].append(file_path)

            # Write output of function to file path if 'force_write' is True.
            if force_write:
                logger.warning(f"Forcing write to file path: {hash_file_path}")
                graph.serialize(hash_file_path)
                results["wrote"].append(hash_file_path)

    # Output results of test.
    # -----------------------

    # Log warning if any expected files were not found.
    if len(results["not_found"]) > 0:
        logger.warning(
            f"\n\n({len(results['not_found'])}) Could not find corresponding hash "
            "files:\n-- " + "\n-- ".join(results["not_found"])
        )

    # Print successfully verified files.
    if len(results["successful"]) > 0:

        # Log final success message.
        logger.info(
            f"\n({len(results['successful'])}) Successfully verified hash methods "
            "against example files:\n-- " + "\n-- ".join(results["successful"])
        )
    else:
        # No files were verified.
        logger.info("No files were verified.")

    if len(results["wrote"]) > 0:
        logger.info(
            f"\n({len(results['wrote'])}) The following files were written to: \n-- "
            + "\n-- ".join(results["wrote"])
        )

    # Fail test if any files failed.
    if len(results["failed"]) > 0:
        pytest.fail(
            f"\n({len(results['failed'])}) The following files failed the hash test: \n-- "
            + "\n-- ".join(results["failed"])
        )


# def test__reverse_hash():
#     test_files = listdir(examples_path)



import argparse
import sys
import logging

from .hash import rdfhash, reverse_hash
from .logger import logger
from .helper import hashlib_methods


def add_io_args(parser: argparse.ArgumentParser):
    """Adds 'input/output' arguments to given parser/subparser.
    Args: parser (argparse.ArgumentParser): parser/subparser
    """
    parser.add_argument(
        "-d", "--data", help="Input RDF string or file. (.ttl, .nt, .n3)"
    )

    parser.add_argument(
        "-f",
        "--format",
        help="Input format.",
        default=None,
        choices=["turtle", "n-triples", "trig", "n-quads", "n3", "rdf"],
    )

    parser.add_argument(
        "-a",
        "--accept",
        help="Output accept format.",
        default=["turtle"],
        nargs="+",
    )


def add_debug_args(parser: argparse.ArgumentParser):
    """Adds 'debug' arguments to given parser/subparser.
    Args: parser (argparse.ArgumentParser): parser/subparser
    """
    parser.add_argument(
        "-v",
        "--verbose",
        help="Show 'info' level logs.",
        action="store_true",
    )

    parser.add_argument(
        "--debug",
        help="Show 'debug' level logs.",
        action="store_true",
    )


def get_parser() -> argparse.ArgumentParser:
    """Return argument parser for command 'rdfhash'.
    Returns: argparse.ArgumentParser: _description_
    """
    parser = argparse.ArgumentParser(
        description="Replace selected subjects with hash of their triples "
        "(`{predicate} {object}.\\n` sorted + joined).",
    )
    subparser = parser.add_subparsers(dest="command")

    # ____________________________
    #
    # Default arguments: 'rdfhash'
    # ____________________________

    add_io_args(parser)

    add_debug_args(parser)

    parser.add_argument(
        "-m",
        "--method",
        help="Hash method.",
        default="sha256",
        choices=hashlib_methods.keys(),
    )

    parser.add_argument(
        "--sparql",
        "--sparql-select-subjects",
        default="SELECT DISTINCT ?s WHERE { ?s ?p ?o . FILTER (isBlank(?s)) }",
        help="SPARQL SELECT query returning subject URIs to replace with hash of"
        " their triples. Defaults to all blank node subjects.",
    )

    # ______________________________________
    #
    # Sub-command 'reverse' ('reverse_hash')
    # ______________________________________

    parser_reverse = subparser.add_parser(
        "reverse", help="Reverse hashed URIs to blank nodes."
    )

    add_io_args(parser_reverse)

    add_debug_args(parser_reverse)

    return parser


def run():
    """
    Parse arguments and pass to function 'rdfhash'. Serialize results with
    respect to 'accept' argument.
    """
    # Parse arguments.
    args_list = sys.argv[1:]
    parser = get_parser()
    args = parser.parse_args(["--help"] if len(args_list) == 0 else sys.argv[1:])

    if args.data == None:
        parser.print_usage()
        print("\nERROR: The following arguments are required: -d/--data")
        sys.exit(1)

    if args.debug:
        logger.setLevel(logging.DEBUG)
    elif args.verbose:
        logger.setLevel(logging.INFO)

    match args.command:
        case None:
            graph = rdfhash(args.data, args.format, args.method, args.sparql)
            print(graph.serialize(format=args.accept[0]))
            sys.exit(0)
        case "reverse":
            graph = reverse_hash(args.data, args.format)
            print(graph.serialize(format=args.accept[0]))
            sys.exit(0)
        case _:
            parser.print_usage()
            print(f"\nERROR: Command is not implemented: {args.command}")
            sys.exit(1)


if __name__ == "__main__":
    run()
